package cytoPrefuse;

import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import prefuse.data.tuple.TupleSet;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class EdgeBundler {
	private TupleSet edges;
	
	private float beta;
	private Point2D p0 = new Point2D.Float();			
    private	Point2D pNminus1 = new Point2D.Float();
   
    private List sortedNodes = new ArrayList();
    
    private static float[] segmentPoints = new float[6];
    private static Line2D line = new Line2D.Float();
    
    public static final int SPLINE_SEGMENT_START = 0;
    public static final int SPLINE_SEGMENT_END = 100;
    public static final int SPLINE_SEGMENT_SIZE = 5;
    public static final int SPLINE_SEGEMENT_COUNT = ((SPLINE_SEGMENT_END - SPLINE_SEGMENT_START)/SPLINE_SEGMENT_SIZE) + 1;

	public EdgeBundler (TupleSet nodes, TupleSet edges){	
		this.edges = edges;
		beta = 1.0f;
		
		Iterator nodesIt = nodes.tuples();
		while (nodesIt.hasNext()) {
			NodeItem node = (NodeItem)nodesIt.next();
			sortedNodes.add(node);
		}
		Collections.sort(sortedNodes, new NodeSorter());
		
	}
	public void setBeta(float beta){
		this.beta = beta;
		//recomputeLocations();
		
		Iterator edgesIt = edges.tuples();
	   while (edgesIt.hasNext()) {
		   EdgeItem edge = (EdgeItem)edgesIt.next();
		   List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
		   betaAdjustControlPoints(controlPoints, (List) edge.get(CytoPrefuse.SPLINE_BETA_ADJUSTED ),controlPoints.size());
		   assignEdgeFromControlPoints(edge);
	   }   
	
	}
	
	public static boolean edgeIntersectsRect(VisualItem edge, Rectangle2D rectangle) {
	   	GeneralPath gp = (GeneralPath)edge.get(CytoPrefuse.EDGE_SPLINE);
			
    	PathIterator pathIt = gp.getPathIterator(new AffineTransform());
    	int segType = pathIt.currentSegment(segmentPoints);
    	float prevX = 0;
    	float prevY = 0;
    	if (segType == PathIterator.SEG_MOVETO){
    		prevX = segmentPoints[0];
    		prevY = segmentPoints[1];
    	}
    	while (!pathIt.isDone()) {    		
    		pathIt.next();
    		segType = pathIt.currentSegment(segmentPoints);
    		if (segType == PathIterator.SEG_LINETO){
    			float curX = segmentPoints[0];
    			float curY = segmentPoints[1];
    			line.setLine(prevX, prevY, curX, curY);
    			if (line.intersects(rectangle)){
    				return true;
    			}
    			prevX = curX;
        		prevY = curY;
        	}else {
    			break;
    		}
    	}
    	return false;
  }
	
	
	public void recomputeLocations(){	    
	    highDegreeNodeCluster();		
	}
		   
	private void highDegreeNodeCluster() {
		Iterator edgeIt = edges.tuples();
		while (edgeIt.hasNext()){
			EdgeItem edge = (EdgeItem)edgeIt.next();
			List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
			controlPoints.clear();
		}
		Iterator nodesIt = sortedNodes.iterator();
		while (nodesIt.hasNext()){
			NodeItem focusNode = (NodeItem)nodesIt.next();
			List edgeClusters = (List)focusNode.get(CytoPrefuse.EDGE_CLUSTERS);
			edgeClusters.clear();
		}
		
		
		Set groupedEdges = new HashSet();	
		
		//Pass 1
		int maxClusterSeparation = 5;
		for (int i=1; i <=maxClusterSeparation; i++){
			int maxSepThisIteration = i * SearchGridLayout.SCALE;
			nodesIt = sortedNodes.iterator();
			while (nodesIt.hasNext()){
				NodeItem focusNode = (NodeItem)nodesIt.next();				
				List edgeClusters = (List)focusNode.get(CytoPrefuse.EDGE_CLUSTERS);
					
				Set focusNodeEdges = new HashSet();
				Iterator edgesIt = focusNode.edges();
				while (edgesIt.hasNext()){
					EdgeItem edge = (EdgeItem)edgesIt.next();
					if (!groupedEdges.contains(edge))	{
						if (getEdgeLength(edge) > i * SearchGridLayout.SCALE ) {							
							focusNodeEdges.add(edge);
						}
					}
				}
				while (!focusNodeEdges.isEmpty()){
					EdgeCluster edgeCluster = new EdgeCluster(focusNode);
					
					EdgeItem edge = (EdgeItem) focusNodeEdges.iterator().next();				
					focusNodeEdges.remove(edge);					
					edgeCluster.addEdge(edge);
					
					List neighboursToCheck = new ArrayList();
					neighboursToCheck.add( edge.getAdjacentNode(focusNode));
			
					while (!neighboursToCheck.isEmpty()){
						NodeItem neighbourNode = (NodeItem)neighboursToCheck.remove(0);
						edgesIt = focusNodeEdges.iterator();
						while (edgesIt.hasNext()){
							EdgeItem nextEdge = (EdgeItem)edgesIt.next();
							NodeItem potentialNeighbour = nextEdge.getAdjacentItem(focusNode);
							double xSep = Math.abs(neighbourNode.getX() - potentialNeighbour.getX());
							double ySep = Math.abs(neighbourNode.getY() - potentialNeighbour.getY());
							if ( (xSep <= maxSepThisIteration && ySep  <= maxSepThisIteration)){
									edgeCluster.addEdge(nextEdge);
									neighboursToCheck.add(potentialNeighbour);
									edgesIt.remove();
							}
						}
					}
				
					if (edgeCluster.size() > 1 || i ==maxClusterSeparation){
						edgeClusters.add(edgeCluster);
						Iterator clusterEdgesIt = edgeCluster.iterator();
						while (clusterEdgesIt.hasNext()){
							EdgeItem e = (EdgeItem)clusterEdgesIt.next();
							groupedEdges.add(e);
						}
					}
				}
			}
		}

		//add the single edges into their own cluster.
		nodesIt = sortedNodes.iterator();
		while (nodesIt.hasNext()){
			NodeItem focusNode = (NodeItem)nodesIt.next();		
			List edgeClusters = (List)focusNode.get(CytoPrefuse.EDGE_CLUSTERS);
		
			Iterator edgesIt = focusNode.edges();
			while (edgesIt.hasNext()){
				EdgeItem edge = (EdgeItem)edgesIt.next();
				if (!groupedEdges.contains(edge))	{
					EdgeCluster single = new EdgeCluster(focusNode);
					single.addEdge(edge);
					edgeClusters.add(single);
					groupedEdges.add(edge);
				}
			}
		}
		
		//Pass 2
		// Now that the edges are clumped into clusters, try joining neighbouring
		// clusters.
	
		NeighbourSorter neighbourSorter = new NeighbourSorter();
	
		
		float angleLimit = (float) (Math.PI/6);
		nodesIt = sortedNodes.iterator();
		while (nodesIt.hasNext()){
			NodeItem focusNode = (NodeItem)nodesIt.next();
			
			List edgeClusters = (List)focusNode.get(CytoPrefuse.EDGE_CLUSTERS);
			
	//		System.out.println(focusNode.getString(CytoPrefuse.LABEL) + ": " + edgeClusters.size());
			
			if (edgeClusters.size() > 0) {
				Collections.sort(edgeClusters, neighbourSorter);			
				
				//Start with each cluster in its own list
				List allClusters = new ArrayList();
				for (int i=0; i < edgeClusters.size(); i++){
					List single = new ArrayList();
					single.add(edgeClusters.get(i));
					allClusters.add (single);
				}
				
				//and merge neighbouring lists until a limit is reached
				List newClusterList = new ArrayList();
				
				boolean changed = false;
				do{
					newClusterList.clear();
					changed = false;
					for (int i=0; i < allClusters.size()-1; i+=2){
						List  group1 = (List) allClusters.get(i);
						List  group2 = (List) allClusters.get(i+1);
						
						float g1angle = getAverageAngle(group1, focusNode);
						float g2angle = getAverageAngle(group2, focusNode);
						
							
						if	 (Math.abs(g1angle - g2angle ) <angleLimit){
							group2.addAll(group1);
							newClusterList.add(group2);
							changed = true;
						}else{
							newClusterList.add(group1);
							newClusterList.add(group2);
						}
					}
					if (allClusters.size()%2 == 1){
						newClusterList.add(allClusters.get(allClusters.size()-1));
					}
					if (!changed && newClusterList.size() > 1){
						//see if the first and last can merge
						List  group1 = (List) newClusterList.get(0);
						List  group2 = (List)newClusterList.get(newClusterList.size()-1);
						
						float g1angle = getAverageAngle(group1, focusNode);
						float g2angle = getAverageAngle(group2, focusNode);
						
						if	 (Math.abs( g1angle - g2angle + (2*Math.PI) ) <angleLimit){
							group2.addAll(group1);
							newClusterList.remove(group1);
							changed = true;
						}
					}
					if (changed){
						
						List temp = allClusters;
						allClusters = newClusterList;
						newClusterList = temp;
					}
				}while (changed);
				
				for (int i=0; i < allClusters.size(); i++){
					setGroupControlPoints((List)allClusters.get(i), focusNode);
				}
			}
		}
		 	
		
		orientEdges();
		
	    edgeIt = edges.tuples();
		
	    while (edgeIt.hasNext()){
		    EdgeItem edge = (EdgeItem)edgeIt.next();
			List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
			fixupSelfEdges(controlPoints);
			betaAdjustControlPoints(controlPoints, (List) edge.get(CytoPrefuse.SPLINE_BETA_ADJUSTED ),controlPoints.size());				
		    assignEdgeFromControlPoints(edge);		   
	    }
	}
	private void orientEdges() {
		//we want the edge to always draw from source to target for colour interpolation.  
		//Some will be reversed to reorient here.
		 Iterator edgeIt = edges.tuples();
		 while (edgeIt.hasNext()){
			    EdgeItem edge = (EdgeItem)edgeIt.next();
				List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
				Point2D lineStart = (Point2D) controlPoints.get(0);
				if ( Math.abs( lineStart.getX() - edge.getSourceItem().getX()) < 0.0001 &&
						Math.abs( lineStart.getY() - edge.getSourceItem().getY()) < 0.0001)
				{
					List reverseControlPoints = new ArrayList();
					for (int i = controlPoints.size()-1; i >=0; i--){
						reverseControlPoints.add(controlPoints.get(i));
					}
					edge.set(CytoPrefuse.SPLINE_CONTROL, reverseControlPoints);
				}
		 }
		
	}
	private void fixupSelfEdges(List controlPoints) {
		Point2D p1 = (Point2D)controlPoints.get(0);
		Point2D pLast = (Point2D)controlPoints.get(controlPoints.size() - 1);
		if (Math.abs(p1.getX() - pLast.getX()) < 0.001 && 
			Math.abs(p1.getY() - pLast.getY()) < 0.001){
			//Self edge.  It's a thin line right now.  Let's make it 
			//a nice curving loop.
			controlPoints.clear();
			controlPoints.add(p1);
			controlPoints.add (new Point2D.Float ((float) (p1.getX() - SearchGridLayout.SCALE/2), (float)p1.getY()));
			controlPoints.add (new Point2D.Float ((float)(p1.getX() - SearchGridLayout.SCALE/2), (float)(p1.getY() - SearchGridLayout.SCALE/2)));
			controlPoints.add (new Point2D.Float ((float)p1.getX(), (float)(p1.getY() - SearchGridLayout.SCALE/2)));
			controlPoints.add (pLast);	
		}
		
	}
	private float getAverageAngle(List group, NodeItem focusNode) 	{
		Iterator ecIt = group.iterator();
		float groupX = 0.0f;
		float groupY = 0.0f;
		float count =0;
		while (ecIt.hasNext()){
			EdgeCluster ec = (EdgeCluster) ecIt.next();
			Point2D p = ec.getEndControlPoint();
			groupX += p.getX() * ec.size();
			groupY += p.getY()* ec.size();
			count+= ec.size();
		}
		return (float) Math.atan2( groupY/count - focusNode.getEndY(), groupX/count - focusNode.getEndX());
	}
	
	private void setGroupControlPoints(List group, NodeItem focusNode) {
		
		if (group.size() ==0)
			return;
		Collections.sort(group, new ClusterEdgeLengthSorter());
		
		Iterator groupIt = group.iterator();
		while (groupIt.hasNext()){
			EdgeCluster cluster = (EdgeCluster) groupIt.next();
			cluster.assignSourceControlPoint();
		}
		
		//Assign a starting control point to show strong clustering leaving the node
		float groupAngle = getAverageAngle(group, focusNode);
		Point2D start = new Point2D.Float((float)((focusNode.getEndX() + (45)* Math.cos(groupAngle))),
				 (float) (focusNode.getEndY() +  (45)*Math.sin(groupAngle)));
		for (int i=0; i < group.size(); i++) {			
			EdgeCluster cluster = (EdgeCluster) group.get(i);
			cluster.addControlPoint(start);
		}
		
		for (int i=0; i < group.size(); i++) {			
			EdgeCluster cluster = (EdgeCluster) group.get(i);
			Point2D p = cluster.getEndControlPoint();
			for (int j=i; j < group.size(); j++) {
				EdgeCluster remainingCluster = (EdgeCluster) group.get(j);
				remainingCluster.addControlPoint(p);
			}
		}	
		
		groupIt = group.iterator();
		while (groupIt.hasNext()){
			EdgeCluster cluster = (EdgeCluster) groupIt.next();
			cluster.assignEndControlPoint();
		}
	}
	

	
	
	private void assignEdgeFromControlPoints(EdgeItem edge) {
		List controlPoints = (List)edge.get(CytoPrefuse.SPLINE_BETA_ADJUSTED);
			
 		GeneralPath path = new GeneralPath();
	 	
		Point2D p1 = (Point2D)controlPoints.get(0);
		path.moveTo((float)p1.getX(), (float)p1.getY());		
		float prevX = (float) p1.getX();
		float prevY = (float) p1.getY();
		int edgeLength = 0;

//debug line version
//		for(int i=0;i<=controlPoints.size() -1;i++){
//		    		Point p = (Point) controlPoints.get(i); 
//		    		path.lineTo(p.x, p.y);
//				    edgeLength += getSegmentLength(prevX, prevY,(int) p.x,(int)p.y);
//				    prevX = (int) p.x;
//				    prevY = (int) p.y;
//		}
//		
//				
		for(int it=SPLINE_SEGMENT_START;it<=SPLINE_SEGMENT_END;it+=SPLINE_SEGMENT_SIZE){
		    Point2D endP = getNextSplinePoint(controlPoints, it);
		    path.lineTo((float)endP.getX(),(float)endP.getY());
		    
		    edgeLength += getSegmentLength(prevX, prevY,(float) endP.getX(),(float)endP.getY());
		    prevX = (int) endP.getX();
		    prevY = (int) endP.getY();
	    }
	    
	    edge.set(CytoPrefuse.EDGE_SPLINE, path);
	    edge.setInt(CytoPrefuse.EDGE_LENGTH, edgeLength);
	}
		public static Point2D getNextSplinePoint(List controlPoints, int it) {
			float t = it / (float)(SPLINE_SEGMENT_END - SPLINE_SEGMENT_START);
		    
			Point2D endP = new Point2D.Float(0.0f,0.0f);
			int nMinus1 = controlPoints.size()-1;
			for(int i=0;i<=nMinus1;i++){
				Point2D p = (Point2D) controlPoints.get(i); 
				float bnit=1.0f;
			    for(int k=1;k<=i;k++){
			        bnit=bnit*t*(nMinus1-k+1)/k;
			    }
			    bnit=(float) (bnit*Math.pow(1.0-t,nMinus1-i));
			    endP.setLocation(
			    		endP.getX()+bnit*p.getX(),
			    		endP.getY()+bnit*p.getY()
			    		);
			}
			return endP;
		}

	private float getSegmentLength(float x1, float y1, float x2, float y2) {
		return (float)  Math.sqrt( (x1-x2) * (x1-x2) + (y1-y2) *(y1-y2));
	}
	
	private float getEdgeLength (EdgeItem e) {
		NodeItem n1 = e.getSourceItem();
		NodeItem n2 = e.getTargetItem();
		return getSegmentLength((float)n1.getEndX(), (float)n1.getEndY(), (float) n2.getEndX(), (float)n2.getEndY());
	}
	private void betaAdjustControlPoints(List points, List betaPoints, int pointCount) {
		betaPoints.clear();
		if (pointCount >=2){
			p0.setLocation((Point2D) points.get(0));
			pNminus1.setLocation((Point2D)points.get(pointCount-1));
			
			for (int i=0; i < pointCount; i++){
				Point2D pi = (Point2D)points.get(i);
				float thisXContribution = (float) (pi.getX() *beta);
				float oneminusbeta = 1-beta;
				float distanceAlongSpline  = i/(pointCount-1);
				float deltaX = (float) (pNminus1.getX() - p0.getX());
				
				float pix = thisXContribution;
				pix += oneminusbeta*((p0.getX() +distanceAlongSpline * deltaX));
		
				float thisYContribution = (float) (pi.getY() *beta);
				float deltaY = (float) (pNminus1.getY() - p0.getY());
				
				float piy =thisYContribution;
				piy += oneminusbeta*((p0.getY() +distanceAlongSpline * deltaY));
				
				betaPoints.add(new Point2D.Float(pix, piy));
			}
		}
	}
	
	private class EdgeCluster {
	
		private float groupX = 0;
		private float groupY = 0;
		private List edges = new ArrayList();
		private NodeItem clusterNode;
		private float minEdgeLength;
		public EdgeCluster (NodeItem clusterNode) {
			this.clusterNode = clusterNode;
			minEdgeLength = Float.MAX_VALUE;
		}
		public void assignEndControlPoint() {
			Iterator edgeIt = edges.iterator();
			while (edgeIt.hasNext()) {
				EdgeItem edge = (EdgeItem)edgeIt.next();
				NodeItem endNode = edge.getAdjacentItem(clusterNode);
				List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
				controlPoints.add(new Point2D.Float ((float)endNode.getEndX(), (float) endNode.getEndY()));		
			}	
		}
		public void addControlPoint(Point2D p) {
			Iterator edgeIt = edges.iterator();			
			while (edgeIt.hasNext()) {
				EdgeItem edge = (EdgeItem)edgeIt.next();
				List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
				Point2D copy = new Point2D.Float();
				copy.setLocation(p);
				controlPoints.add(copy);
			}	
					
		}
		public Point2D getEndControlPoint() {			
			int controlDistance;
			if (minEdgeLength > 100 ){
				controlDistance = 45;
			}else {
				controlDistance = 15;
			}
			
			float groupAngle = getClusterAngle();
			return new Point2D.Float((float) (clusterNode.getEndX() + (minEdgeLength-controlDistance)* Math.cos(groupAngle)),
					 (float)((clusterNode.getEndY() +  (minEdgeLength-controlDistance)*Math.sin(groupAngle))));
		}
		
		public void assignSourceControlPoint() {
			Iterator edgeIt = edges.iterator();
			while (edgeIt.hasNext()) {
				EdgeItem edge = (EdgeItem)edgeIt.next();
				List controlPoints = (List) edge.get(CytoPrefuse.SPLINE_CONTROL);
				if (controlPoints.size()!=0) {
					System.out.println("That's very very weird");
				}
				controlPoints.add(new Point2D.Float ((float)clusterNode.getEndX(), (float) clusterNode.getEndY()));		
			}	
		}
		
		void addEdge(EdgeItem e){
			edges.add(e);
			NodeItem n1 = e.getAdjacentItem(clusterNode);
			groupX += n1.getEndX() - clusterNode.getEndX() ;
			groupY += n1.getEndY() - clusterNode.getEndY();
			float thisEdgeLength = getSegmentLength((float)clusterNode.getEndX(), (float)clusterNode.getEndY(),(float) n1.getEndX(),(float) n1.getEndY());
			if (thisEdgeLength<minEdgeLength){
				minEdgeLength = thisEdgeLength;
			}
		
		}
		void clear() {
			edges.clear();
			groupX = 0;
			groupY = 0;
		}
		int size() {
			return edges.size();
		}
		Iterator iterator() {
			return edges.iterator();
		}
		public float getClusterAngle() {
			float theta =  (float) Math.atan2( groupY/edges.size(), groupX/edges.size());
	//		if (theta < 0) {
	//			theta += Math.PI * 2;
	//		}
			return theta;
		}
		public float getAverageEdgeLength() {	 
			return (float) Math.sqrt( (groupX/edges.size()) * (groupX/edges.size()) + (groupY/edges.size())*(groupY/edges.size()));
		}
	}
	
	private class ClusterEdgeLengthSorter implements Comparator {
		public int compare(Object arg0, Object arg1) {
			EdgeCluster ec1 = (EdgeCluster)arg0;
			EdgeCluster ec2 = (EdgeCluster)arg1;
			if (ec1.getAverageEdgeLength() < ec2.getAverageEdgeLength()){
				return -1;
			}else{
				return 1;
			}
		}
	}
	
	private class NodeSorter implements Comparator {

		public int compare(Object arg0, Object arg1) {
			NodeItem n1 = (NodeItem)arg0;
			NodeItem n2 = (NodeItem)arg1;
			if (n1.getDegree() < n2.getDegree())
				return 1;
			if (n1.getDegree() == n2.getDegree())
				return 0;
			else
				return -1;
		}
	}
	
	private class NeighbourSorter implements Comparator {
		
		public int compare(Object arg0, Object arg1) {
			if (arg1 == null)
				return -1;
			if (arg0 == null)
				return 1;
			EdgeCluster edgeCluster = (EdgeCluster)arg0;
			float theta1 = edgeCluster.getClusterAngle();
			
			edgeCluster = (EdgeCluster)arg1;
			float theta2 = edgeCluster.getClusterAngle();
			
			if (theta1 < theta2) {
				return -1;
			}else 
				return 1;
		}
	 }
}
