/**
 * 
 */
package cytoPrefuse;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import prefuse.Display;
import prefuse.action.Action;
import prefuse.data.Node;
import prefuse.render.Renderer;
import prefuse.util.ColorLib;
import prefuse.util.FontLib;
import prefuse.util.PrefuseLib;
import prefuse.util.display.PaintListener;
import prefuse.visual.AggregateItem;
import prefuse.visual.DecoratorItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class LabelLayout extends Action  implements PaintListener {
  private int noScaleFontSize =12;
   private DecoratorItem[] orderedNodes; 
   private int xMax;
   private int yMax;
   private int lastFontSize =0;
   private boolean requiresRepaint;
   private boolean [][]freeSpace;
   private Comparator sorter;
	private Map functionGroups = new HashMap();
	private float extraSpaceX;
	private float extraSpaceY;
	private CytoPrefuse cytoPrefuse;
	private Display d;
	private Rectangle2D r = new Rectangle2D.Double();
	private final int GRIDIFY = SearchGridLayout.SCALE /2;
	private float aggregateScale;
	private boolean renderAggregateLabels = false;
	
	public LabelLayout( DecoratorItem[] labelOrder, Comparator sorter, CytoPrefuse cytoPrefuse) {
		orderedNodes = labelOrder;
		this.sorter = sorter;
		this.cytoPrefuse = cytoPrefuse;
		this.d = (Display)cytoPrefuse;
		
	   cytoPrefuse.getVisualization().getBounds( CytoPrefuse.NODES,r);
	   xMax = (int) (r.getWidth() / GRIDIFY) + 5;
	   yMax = (int) (r.getHeight() / GRIDIFY) + 5;
	   freeSpace = new boolean[xMax][yMax];
	   requiresRepaint = false;
	   setAggregateLabelScale(1.0f);
	   for (int i=0; i < orderedNodes.length; i++){
		   DecoratorItem item = orderedNodes[i];
		   Object o = item.getDecoratedItem();
		   if (o instanceof NodeItem ){
	          VisualItem node = item.getDecoratedItem();
		       if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
		          List functionNeighbours = (List) node.get(CytoPrefuse.FUNCTION_NEIGHBOURS);
		          
		          if (functionNeighbours != null){
		        	  functionGroups.put(new Integer(node.getInt(CytoPrefuse.BIO_FUNCTION)), functionNeighbours);
		          }
		      }
		   }
	   }
	}
	
	public void setAggregateLabelScale(float scale){
		aggregateScale =scale;
		if (scale > 0.01){
		     renderAggregateLabels = true;
		}else{
			renderAggregateLabels =false;
		}
		setLabelPosition();
		d.repaint();
	}
	public void setExtraSpace (float extraSpace){
		extraSpaceX  = extraSpace * xMax;
		extraSpaceY = extraSpace * yMax;
		setLabelPosition();
		d.repaint();
	}

	public void setLabelPosition() {
          
    	Font defaultFont;

 	   cytoPrefuse.getVisualization().getBounds( CytoPrefuse.NODES,r);
 	   int curXMax = (int) (r.getWidth() / GRIDIFY) +5;
	   int curYMax = (int) (r.getHeight() / GRIDIFY) +5;

    	if (xMax !=curXMax ||
    			yMax != curYMax){
    		xMax = curXMax;
    		yMax = curYMax;

    		freeSpace = new boolean[xMax][yMax];
    	}
    	
    	
    	int fontSize = (int) (noScaleFontSize * (1/d.getScale()));
        
    	if (fontSize < noScaleFontSize){
    		fontSize = noScaleFontSize;
    	}
       
       	defaultFont =FontLib.getFont("Tahoma",fontSize);
   	 
   	    Color bgColor= d.getBackground();   	   
    	 Color transparentBG = ColorLib.getColor(bgColor.getRed(), bgColor.getGreen(), bgColor.getBlue(), 150);
    	    
    	 
    	 for (int i=0; i < xMax; i++){
    		 Arrays.fill(freeSpace[i], true);
    	 }
    	 

         
    	 
    	 for (int i=0; i < orderedNodes.length; i++){
            DecoratorItem item = orderedNodes[i];
            item.setFillColor(ColorLib.color(transparentBG));
        	          
            Object o = item.getDecoratedItem();
    	 	
	         if (o instanceof NodeItem ){
	            VisualItem node = (VisualItem)o;
	            
	    		String label = node.getString(CytoPrefuse.LABEL);	    		
	    		if (label == null || label.equalsIgnoreCase("") || !node.isVisible()){
	    			item.setVisible(false);
	    		}else{
					
		            Font baseFont = (Font) node.get(CytoPrefuse.BASE_FONT);
		            Font font = null;
		            if (baseFont !=null){
		            	font =FontLib.getFont(baseFont.getFontName(),baseFont.getSize()* (1/d.getScale()));
		            }else {
		               	font =FontLib.getFont(defaultFont.getFontName(),defaultFont.getSize()* (1/d.getScale()));
		   		    }
		        	FontMetrics fm = Renderer.DEFAULT_GRAPHICS.getFontMetrics(font);
	   
		            PrefuseLib.setX(item, null, node.getX());
		            PrefuseLib.setY(item, null, node.getY() + (node.getSize() * 5) + fm.getHeight());
		            
		            tryPlaceLabel(font, fm, item, false);
		            item.setTextColor(cytoPrefuse.colorManager.labelColorAction.getColor(item));
		            item.setFillColor(cytoPrefuse.colorManager.labelBGColorAction.getColor(item));
	    		}
		   }else {
	        	//Aggregate label
			   	if (renderAggregateLabels){
	             	//display function group name for groups
	            	AggregateItem aggItem = (AggregateItem)o;
				    if (aggItem.isVisible()){
	             		Integer groupID = new Integer(item.getInt(CytoPrefuse.BIO_FUNCTION));
	             		List groupMembers = (List) functionGroups.get(groupID);
	             		if (groupMembers != null){
	             			 item.setVisible(true);
	             			              	        	
	                         double x = 0.0;
	                         double centerY = 0.0;
	                         int minY = Integer.MAX_VALUE;
	                         Iterator members = groupMembers.iterator();
	                         boolean allInvisible = true;
	                         while (members.hasNext()){
	                         	 NodeItem member = (NodeItem) members.next();
	                         	 x += member.getX();
	                         	 centerY += member.getY();
	                         	 if (member.getY() < minY)
	                         		 minY = (int)member.getY();
	                         	 if (member.isVisible())
	                         		 allInvisible = false;
	                         }
	                         centerY /= groupMembers.size();
	                          
	                         PrefuseLib.setX(item, null, x/groupMembers.size());
	                      
	                         NodeItem aMember = (NodeItem) groupMembers.iterator().next();
	                         Rectangle2D renderBounds = 	aMember.getBounds();                     	                          
	                 		   double y = minY -   renderBounds.getHeight()/2;
	                 		 
	                 		   double groupCenteringFactor = 1.0 - d.getScale();
	                 		   if (groupCenteringFactor < 0)
	                 			  groupCenteringFactor = 0;
	                 		   y += groupCenteringFactor * (centerY-y);
	                 		 
	                         PrefuseLib.setY(item, null, y);
	                         item.setTextColor(ColorLib.gray(0));
	                         
	                         Font baseFont = (Font) aMember.get(CytoPrefuse.BASE_FONT);
	     		             Font font = null;
	     		             if (baseFont !=null){
	     		            	font =FontLib.getFont(baseFont.getFontName(),  noScaleFontSize *aggregateScale * (1/d.getScale()));
	     		             }else {
	     		            	font = FontLib.getFont("Tahoma",  noScaleFontSize *aggregateScale * (1/d.getScale()));
	     		            }
	     		          	    
	                     	  FontMetrics fm = Renderer.DEFAULT_GRAPHICS.getFontMetrics(font);
	                        
	                     	 if (allInvisible)
	                     		 item.setVisible(false);
	                     	 else
	                     		 tryPlaceLabel(font, fm, item, false);	     
	             		}  else{
	             			item.setVisible(false);
	             		}	             		
		        } else{
		        	item.setVisible(false);
		         }
			 	}else{
			   		item.setVisible(false);
			  	}
		   }
    	 }
    	 

        
    }

	private void tryPlaceLabel(Font font, FontMetrics fm, DecoratorItem item, boolean ignoreSpaceConstraints) {
		boolean spaceEmpty = true;
		item.setFont(font);
		Rectangle2D renderBounds = 	item.getBounds();
			
		int xLow = (int) ((renderBounds.getMinX() / GRIDIFY) - extraSpaceX);
		if (xLow < 0)
			xLow =0;
		if (xLow > xMax -1)
			xLow = xMax -1;
        
		int xHigh = (int) (( renderBounds.getMaxX() / GRIDIFY) + extraSpaceX);
		if (xHigh > xMax -1)
			xHigh = xMax -1;
		if (xHigh < 0)
			xHigh = 0;
	
		int yLow = (int) ((renderBounds.getMinY() /GRIDIFY) - extraSpaceY);			
		if (yLow < 0)
			yLow = 0;
		if (yLow > yMax -1)
			yLow = yMax - 1;

		int yHigh = (int) ((renderBounds.getMaxY() /GRIDIFY) + extraSpaceY);
		if (yHigh >  yMax -1)
			yHigh = yMax -1;
		if (yHigh < 0)
			yHigh = 0;
		
		if (!ignoreSpaceConstraints){
			for (int x=xLow; x <=xHigh; x++){
				for (int y= yLow; y <=yHigh; y++) {
					if (!freeSpace [ x][y]){
						spaceEmpty=false;
						break;
					}
				}
				if (!spaceEmpty){
					break;
				}
			}
		}
		if (spaceEmpty || ignoreSpaceConstraints){
			if (!ignoreSpaceConstraints){
					for (int x=xLow; x <=xHigh; x++){
						for (int y= yLow; y <=yHigh; y++) {
				    		freeSpace [ x][y] = false;		                											
						}						
					}
			}
		    item.setVisible(true);
		    
		}
		else{
			item.setVisible(false);
		}
	}
 
	public void prePaint(Display d, Graphics2D g) {
		
		//see if the sort still holds.
		for (int i=0; i < orderedNodes.length - 1; i++){
			if (sorter.compare(orderedNodes[i], orderedNodes[i+1]) == 1){
				requiresRepaint = true;
				Arrays.sort(orderedNodes, sorter);
				//Indicate we need a full screen repaint as our label positions will change.
				d.setDamageRedraw(false);
				break;
			}
		}
		int fontSize = (int) (noScaleFontSize * (1/d.getScale()));
		if (fontSize != lastFontSize || requiresRepaint){
			requiresRepaint = false;
			lastFontSize = fontSize;
			setLabelPosition();
		}
	
				
	}

	public void postPaint(Display d, Graphics2D g) {
		d.setDamageRedraw(true);
	}

	public void run(double frac) {
		setLabelPosition();
	}


}