package prefuse.controls;

import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import prefuse.Visualization;
import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.expression.AndPredicate;
import prefuse.data.expression.BooleanLiteral;
import prefuse.data.expression.ColumnExpression;
import prefuse.data.expression.ComparisonPredicate;
import prefuse.data.expression.Predicate;
import prefuse.data.tuple.TupleSet;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualGraph;
import prefuse.visual.VisualItem;
import cytoPrefuse.CytoPrefuse;

public class PathwayChooser extends ControlAdapter {
	private static final String DIJKSTRA_PREVIOUS = "dijkstra_previous";
	private static final String DIJKSTRA_D = "dijkstraD";
	private VisualItem begin;
	private VisualItem end;
	private VisualGraph g;
	private List pathNodes;
	private List pathEdges;
	private Visualization vis;
	private Predicate onPath;
	private Predicate onPathNotOnShortest;
	private boolean showingPath;
	private Set S = new HashSet();
	private List Q = new ArrayList();
	private CytoPrefuse cytoPrefuse;
	
	public PathwayChooser (VisualGraph graph, Visualization inVis) {
		vis = inVis;
		g = graph;
		cytoPrefuse = (CytoPrefuse)inVis.getDisplay(0);
		
		g.addColumn(DIJKSTRA_D, int.class);
		g.addColumn(DIJKSTRA_PREVIOUS, NodeItem.class);

		pathNodes = new ArrayList();
		pathEdges = new ArrayList();
		onPath =  new ComparisonPredicate(ComparisonPredicate.EQ, new ColumnExpression(CytoPrefuse.ON_PATHWAY), new BooleanLiteral(true));
		onPathNotOnShortest =new AndPredicate(  new ComparisonPredicate(ComparisonPredicate.EQ, new ColumnExpression(CytoPrefuse.ON_PATHWAY), new BooleanLiteral(true)),
																				new ComparisonPredicate(ComparisonPredicate.EQ, new ColumnExpression(CytoPrefuse.ON_SHORTEST_PATH), new BooleanLiteral(false)));
			
		showingPath = false;
	}
	public void itemEntered(VisualItem item, MouseEvent e) {
	    if ( item instanceof NodeItem ){
	        if (showingPath){
	        	if (item.getBoolean(CytoPrefuse.ON_PATHWAY)){
	        		clearOldPathway(true);
	        		List halfPath = findShortestPath(g, (NodeItem)begin, (NodeItem)item, (NodeItem)end,  onPath );
	        		assignPathAttributes(halfPath, CytoPrefuse.ON_SHORTEST_PATH);            	
	        	
	        		List secondHalfPath = findShortestPath(g, (NodeItem)item, (NodeItem)end, null, onPathNotOnShortest);   
	        
	        		if (secondHalfPath.size() > 0){
	        			assignPathAttributes(secondHalfPath, CytoPrefuse.ON_SHORTEST_PATH);
	        		}else{
	        			//couldn't find a way through.  Must be on a cycle.  
	        			//Ok, find the path but this time allow visiting existing path nodes
	        			 secondHalfPath = findShortestPath(g, (NodeItem)item, (NodeItem)end, null, onPath);   
	        			 //find the point where it rejoins the shortes path
	        			 NodeItem rejoinNode = null;
	        			 Iterator pathIt = secondHalfPath.iterator();
	        			 while (pathIt.hasNext()){
	        				NodeItem nodeItem = (NodeItem) pathIt.next();
	        				if (nodeItem.getBoolean(CytoPrefuse.ON_SHORTEST_PATH)){
	        					rejoinNode = nodeItem;
	        					break;
	        				}
	        			 }
	        			 //Assign the shortest path
	        			 assignPathAttributes(secondHalfPath, CytoPrefuse.ON_SHORTEST_PATH);
	        		     
	        			 //And complete the cycle by finding the shortest path back to the pathway
	        			 List completeCycle = findShortestPath(g, (NodeItem)item, rejoinNode, null, onPathNotOnShortest);
	        			 assignPathAttributes(completeCycle, CytoPrefuse.ON_SHORTEST_PATH);	        			 
	        		}
	        		
	        		vis.run("color");
	        	}
	        }
		 }  
	 }
	public void itemExited(VisualItem item, MouseEvent e) {
	 }
	 
	 public void itemClicked(VisualItem item,
             java.awt.event.MouseEvent e) {
		if ( item instanceof NodeItem ){
			 if ((e.getModifiers() & InputEvent.CTRL_MASK) == InputEvent.CTRL_MASK) {
				 if (showingPath || begin == null) {
					cytoPrefuse.selectInCytoscape(pathNodes.toArray(), false);
					cytoPrefuse.selectInCytoscape(pathEdges.toArray(), false);

					 clearOldPathway(false);
					 showingPath = false;
					 begin = item;
					 begin.setBoolean(CytoPrefuse.ON_SHORTEST_PATH, true);				
					 cytoPrefuse.colorManager.pathColoring = false;
				 }
				 else {
					 if (begin.getRow() == item.getRow()) {
						 //clicking again to deselect
						 begin.setBoolean(CytoPrefuse.ON_SHORTEST_PATH, false);
						 begin = null;
					 }else {
						 //clicking the end point
						 end = item;
						 end.setBoolean(CytoPrefuse.ON_SHORTEST_PATH, true);
						 createPathway(g);				 
						 showingPath = true;
					 }
				 } 
				 vis.run("color");
					 
			 }
		 }
	 }
	private void clearOldPathway(boolean clearShortestPathOnly) {
		//			Clear out the old pathway
		Iterator pathIt = pathNodes.iterator();
		while (pathIt.hasNext()){
			NodeItem node = (NodeItem) pathIt.next();
			if (!clearShortestPathOnly)
				node.setBoolean(CytoPrefuse.ON_PATHWAY, false);
			node.setBoolean(CytoPrefuse.ON_SHORTEST_PATH, false);
		}
		if (!clearShortestPathOnly)			
			pathNodes.clear();
		pathIt = pathEdges.iterator();
		while (pathIt.hasNext()) {
			Edge edge = (Edge) pathIt.next();
			if (!clearShortestPathOnly)				
				edge.setBoolean(CytoPrefuse.ON_PATHWAY, false);
			edge.setBoolean(CytoPrefuse.ON_SHORTEST_PATH, false);
		}
		if (!clearShortestPathOnly)			
			pathEdges.clear();					
	}
	private void createPathway(VisualGraph g) {
		

		//If the begin row is a higher level than the end row, swap roles
		int beginLevel = begin.getInt(CytoPrefuse.BIO_CATEGORY);
		int endLevel = end.getInt(CytoPrefuse.BIO_CATEGORY);
		if (endLevel < beginLevel && !end.getBoolean( CytoPrefuse.ANY_LAYER)) {
			VisualItem temp = begin;
			begin = end;
			end = temp;
		}
		
		//Use Dijkstra's algorithm to find the
		//shortest path from begin to end		
	
		int pathCount =0;			
		List p = findShortestPath(g, (NodeItem)begin, (NodeItem)end, null, null);
		pathNodes.addAll(p);		
		assignPathAttributes(pathNodes,CytoPrefuse.ON_SHORTEST_PATH);
		pathCount += pathNodes.size();
		
		Set nodesOnPath = new HashSet();
		List nodesToCheckLater = new LinkedList();
		findPossiblePaths ((NodeItem)begin, (NodeItem)end,null,  new HashSet(), new HashSet(), nodesOnPath, nodesToCheckLater);
	
		Iterator checkLaterIt = nodesToCheckLater.iterator();
		while (checkLaterIt.hasNext()) {
			List possibleList = (List) checkLaterIt.next();
			NodeItem head = (NodeItem) possibleList.get(0);
			if (nodesOnPath.contains(head)){
				nodesOnPath.addAll(possibleList);
			}
		}
		nodesOnPath.add(end);
		nodesOnPath.add(begin);
		TupleSet focusItems = vis.getGroup(CytoPrefuse.SELECTED);
		Iterator pathIt = nodesOnPath.iterator();
		while (pathIt.hasNext()) {
			NodeItem n = (NodeItem) pathIt.next();
			n.setBoolean(CytoPrefuse.ON_PATHWAY, true);
			focusItems.addTuple(n);
			
		}
		pathCount += nodesOnPath.size();
		pathNodes.addAll(nodesOnPath);
		
		Iterator edgesIt = g.edges();
		while (edgesIt.hasNext()) {
			Edge e = (Edge) edgesIt.next();
				
			if (e.getSourceNode().getBoolean(CytoPrefuse.ON_PATHWAY) || 
					e.getSourceNode().getBoolean(CytoPrefuse.ON_SHORTEST_PATH)){
				if (e.getTargetNode().getBoolean(CytoPrefuse.ON_SHORTEST_PATH)||
						e.getTargetNode().getBoolean(CytoPrefuse.ON_PATHWAY) ){
					e.setBoolean(CytoPrefuse.ON_PATHWAY, true);
					pathEdges.add(e);
				}
			}
		}
		cytoPrefuse.colorManager.pathColoring = true;
		
	}
	private void findPossiblePaths(NodeItem n, NodeItem searchTarget,Edge inEdge, Set candidateNodes, Set nodesVisited, Set nodesOnPath, List nodesToCheckLater) {
		candidateNodes.add(n);
		nodesVisited.add(n);
		Iterator edgesIt = n.edges();
		while (edgesIt.hasNext()) {
			Edge e = (Edge) edgesIt.next();
			if (e != inEdge) {
				NodeItem target = (NodeItem) e.getAdjacentNode(n);
				if (validPathwayNeighbour(n, target)){
					if (target == searchTarget) {
						nodesOnPath.addAll(candidateNodes);
						candidateNodes.clear();
					}else {
					
						if (!nodesVisited.contains(target)){
							findPossiblePaths(target, searchTarget, e, candidateNodes, nodesVisited, nodesOnPath, nodesToCheckLater);
						}else {
							if (nodesOnPath.contains(target)) {
								nodesOnPath.addAll(candidateNodes);
								candidateNodes.clear();
							}else {
								//We've just gone round a cycle, but we're not sure yet
								//whether this node will eventually join the path.  Add it
								//to a list to check later.
								List possibleNodes = new LinkedList();
								possibleNodes.addAll(candidateNodes);
								possibleNodes.add(0, target);
								nodesToCheckLater.add(possibleNodes);
								
							}
						}
					}
				}
			}
		}
		candidateNodes.remove(n);
	}
	private boolean validPathwayNeighbour(NodeItem n, NodeItem target) {
		if (n.getBoolean(CytoPrefuse.ANY_LAYER))
			return true;
		if (target.getBoolean(CytoPrefuse.ANY_LAYER))
			return true;
		
		int nodeLayer = n.getInt(CytoPrefuse.BIO_CATEGORY);
		int targetLayer = target.getInt(CytoPrefuse.BIO_CATEGORY);
		return targetLayer == (nodeLayer + 1) || 
				targetLayer == nodeLayer; 
	}
	
	
	
	private void assignPathAttributes(List path, String pathway) {
		Iterator pathIt = path.iterator();
		NodeItem prevItem = null;
		while (pathIt.hasNext()){
			NodeItem nodeItem = (NodeItem) pathIt.next();
			nodeItem.setBoolean(pathway, true);
			if (prevItem != null){
				Iterator edgesIt = nodeItem.edges();
				while (edgesIt.hasNext()){
					Edge e = (Edge)edgesIt.next();
					if (e.getAdjacentNode(nodeItem).getRow() == prevItem.getRow()) {
						e.setBoolean(pathway, true);
						pathEdges.add(e);
						break;
					}
				}			
			}
			prevItem = nodeItem;
		}
	}
	
	private List findShortestPath(Graph g, NodeItem beginNode, NodeItem endNode, NodeItem nodeToExclude, Predicate filter) {
		Iterator nodeIt = g.getNodes().tuples(filter);
		while (nodeIt.hasNext()){
			NodeItem nodeItem = (NodeItem) nodeIt.next();
			dijkstraInitialize(nodeItem);
		}
		dijkstraInitialize(beginNode);
		dijkstraInitialize(endNode);
	
		
		//Distance to the source is 0
		beginNode.setInt(DIJKSTRA_D,0);
		
		S.clear();
		Q.clear();
		
		nodeIt = g.getNodes().tuples(filter);
		while (nodeIt.hasNext()){
			NodeItem entry = (NodeItem) nodeIt.next();
			if (entry != nodeToExclude)
				Q.add(entry);
		}
		
		//Don't allow filtering out the begin and end node.
		//If they didn't pass the filter, add them in here.
		if (filter != null){
			if (!filter.getBoolean(beginNode))
				Q.add(beginNode);
			if (!filter.getBoolean(endNode))
				Q.add(endNode);
		}
		
		
		while (!Q.isEmpty()){
			//Extract the min element from the q
			int min = Integer.MAX_VALUE;
			NodeItem u = null;
			Iterator qIt = Q.iterator();
			while (qIt.hasNext()) {
				NodeItem qTest = (NodeItem) qIt.next();
				if (qTest.getInt(DIJKSTRA_D) < min) {
					min = qTest.getInt(DIJKSTRA_D);
					u = qTest;
				}
			}
			Q.remove(u);
			if (u.getRow() == endNode.getRow()){
				break;
			}
			S.add(u);
			
			Iterator edgesIt = u.edges();
			int du = u.getInt(DIJKSTRA_D);
			while (edgesIt.hasNext()) {
				Edge e = (Edge) edgesIt.next();
				if (filter!=null){
					if (!filter.getBoolean(e)){
						continue;
					}
				}
				NodeItem target = (NodeItem) e.getAdjacentNode(u);
				if (validPathwayNeighbour(u, target)){
					if (!S.contains(target)){
						int dv = target.getInt(DIJKSTRA_D);
						if (du +1 < dv) {
							dv = du +1;
							target.setInt(DIJKSTRA_D, dv);
							target.set(DIJKSTRA_PREVIOUS, u);
						}
					}
				}
			}
		}
		
		//now extract the path
		List p = new ArrayList();
		
		NodeItem u = endNode;
		NodeItem prev = (NodeItem) u.get(DIJKSTRA_PREVIOUS);
		while (prev != null) {
			p.add(u);
			u = prev;				
			prev = (NodeItem) u.get(DIJKSTRA_PREVIOUS);
		}
		p.add(u);
		if (u != beginNode){
			System.out.println("No path found");
			p.clear();
		}
		return p;
	}
	private void dijkstraInitialize(NodeItem nodeItem) {
		nodeItem.setInt(DIJKSTRA_D, Integer.MAX_VALUE/2);
		nodeItem.set(DIJKSTRA_PREVIOUS, null);
	}
	
	 
}
